# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Constants used for server connections."""

from enum import Enum
from pathlib import Path

LOCALHOST = "127.0.0.1"
"""IP address for localhost."""
PYPIM_PRODUCT_NAME = "additive"
"""Product name for the Additive server in a PyPIM environment."""
DEFAULT_PRODUCT_VERSION = "261"
"""Default Ansys product version to use for the Additive server."""
ADDITIVE_SERVER_EXE_NAME = "additiveserver"
"""Name of the Additive server executable."""
ADDITIVE_SERVER_SUBDIR = Path("Additive") / "additiveserver"
"""Subdirectory for the Additive server in the Ansys installation directory."""
UNIX_DOMAIN_SOCKET_SERVICE_NAME = "additive"
"""Name of the Unix Domain Socket service."""


class TransportMode(str, Enum):
    """Enum containing the different modes of connection."""

    (INSECURE, MTLS, UDS) = ("insecure", "mtls", "uds")
